// display_unit
// The all-in-one video display management unit for this maze game.
// Sam Revitch, CSE 467

// This unit integrates all of the necessary hardware to drive a VGA style
// display, using a set of sprite registers, a custom maze RAM for background,
// and a CPU interface, including a sequential configuration write register,
// access to the maze RAM, and VBLANK IRQs after every frame has finished.

// Description of the configuration memory:
// The configuration memory is written to by loading data into the cfg_write
// input, and strobing cfg_we.  The write occurs at the end of the clock
// cycle when cfg_we goes high.  There are many configuration registers
// behind this interface, and sequential write cycles write to sequentially
// higher indexed registers.  After all registers have been written to,
// writing resumes at the first register.
// Byte    Descr
// 0       Maze foreground color, [5:4]: red, [3:2]: green, [1:0]: blue.
// 1       Maze background color, [5:4]: red, [3:2]: green, [1:0]: blue.
// 2-6     Sprite 0 configuration:
//         2: Image configuration byte, see pacman_sprite_unit below
//         3: [4:0]: bits [9:5] of the X coordinate of the upper-left corner
//         4: [4:0]: bits [4:0] of the X coordinate of the upper-left corner
//         5: [4:0]: bits [9:5] of the Y coordinate of the upper-left corner
//         6: [4:0]: bits [4:0] of the Y coordinate of the upper-left corner
// 7-11    Sprite 1 configuration.
// 12-16   Sprite 2 configuration.
// 17-21   Sprite 3 configuration.

// Description of the maze RAM access ports:
// The Maze RAM can be accessed by feeding an address into the maccess_addr
// input.  maccess_addr is really a bitfield representing:
//
// maccess_addr[7:4]    Row of interest
// maccess_addr[3:0]    Column of interest
//
// Note that cycling through maccess_addr in order results in a regular
// left-to-right, top-to-bottom traversal of the maze.
//
// All access to the maze RAM is delayed until the next positive clock edge,
// and should ideally be initiated with as little setup time as possible.
// see mazekeeper.v for a description of the format of the four read and
// write bitfields.

// Description of the status_read register:
// status_read[0]     true indicates a recent hardware reset, gets reset to 0
//                    following a configuration register write cycle.
// status_read[7:1]   unused, all set to 0.

// Description of positive_vblank:
// This register goes high for one clock cycle immediately following a
// vertical blank.  Useful for starting a device responsible for computing
// new game state.

// Description of h_sync_out, v_sync_out:
// The display unit is internally pipelined.  If it is driven by external
// h_sync and v_sync generators, the pipeline latency must be accounted for
// such to keep the output RGB signals in sync.  To compensate, the sync
// signals are placed into their own pipeline of equal depth.

// Description of rgb_out:
// The display_unit currently outputs video signals in a 6BPP RGB format.
// rgb_out[5] = MSB red
// rgb_out[4] = LSB red
// rgb_out[3] = MSB green
// rgb_out[2] = LSB green
// rgb_out[1] = MSB blue
// rgb_out[0] = LSB blue
//
// rgb_out is latched, fully suitable to be sent directly to the VGA port
// on a proto board, and will exhibit minimal sub-pixel noise.

module display_unit(clk, reset,
		    h_sync, v_sync, h_rgb, v_rgb,
		    maccess_addr, maccess_read, maccess_we, maccess_write,
		    status_read, cfg_write, cfg_we,
		    positive_vblank,
		    h_sync_out, v_sync_out, rgb_out);

   // Display interface
   input clk, reset;
   output h_sync, v_sync, h_rgb, v_rgb;
   output      h_sync_out, v_sync_out;
   output [5:0] rgb_out;

   // Game state controller interface
   input [7:0] 	maccess_addr;
   input [3:0] 	maccess_write, maccess_we;
   output [3:0] maccess_read;
   output [7:0] status_read;
   input [7:0] 	cfg_write;
   input 	cfg_we;
   output 	positive_vblank;

   // horizontal and vertical sync signal generator
   vga_hv_sync_unit local_hv(clk, reset, h_sync, v_sync, h_rgb, v_rgb);

   // Sync signal pipeline/delay
   // NOTE: make sure this is *as deep* as the pixel value pipeline.
   reg [1:0] 	hv_sync0, hv_sync1, hv_sync2, hv_sync3;
   assign h_sync_out = hv_sync3[1];
   assign v_sync_out = hv_sync3[0];

   always @(posedge clk) begin
      hv_sync0 <= {h_sync, v_sync};
      hv_sync1 <= hv_sync0;
      hv_sync2 <= hv_sync1;
      hv_sync3 <= hv_sync2;
   end // always @ (posedge clk)


   // RGB output signal delay
   // Forces a nil RGB signal
   reg rgb_en0, rgb_en1, rgb_en2, rgb_en3;
   wire end_rgb_enable;
   always @(posedge clk) begin
      rgb_en0 <= (h_rgb && v_rgb);
      rgb_en1 <= rgb_en0;
      rgb_en2 <= rgb_en1;
      rgb_en3 <= rgb_en2;
   end
   assign end_rgb_enable = rgb_en3;


   // Screen position counter functionality
   reg [9:0] 	h_pos, v_pos;
   reg [1:0] 	prev_hv_rgb;
   reg 		positive_vblank;
   wire [1:0] 	hv_rgb;
   wire [1:0] 	hv_blank;

   assign hv_rgb = {h_rgb, v_rgb};
   // hv_blank is for texture units below and operates on t-1 time.
   assign hv_blank = { (prev_hv_rgb == 2'b01),
		       (!prev_hv_rgb[0]) };
   
   // horizontal and vertical position drivers   
   always @(posedge clk) begin

      if(!hv_rgb[0] && prev_hv_rgb[0]) begin
//	 $display("Entered V-blank @ (%d,%d)", h_pos, v_pos);
	 positive_vblank <= 1;
      end else begin // if (!hv_rgb[0] && prev_hv_rgb[0])
//	 if(positive_vblank)
//	    $display("left positive vblank");
	 positive_vblank <= 0;
      end

      // If in a V-blank..
      if(!hv_rgb[0]) begin
	 h_pos <= 0;
	 v_pos <= 0;
      end // if (!hv_rgb && prev_hv_rgb[0])

      // Otherwise..
      if(hv_rgb[0]) begin

	 // If entering an H-blank..
	 if(!hv_rgb[1] && prev_hv_rgb[1]) begin
//	    $display("Entered H-blank @ (%d,%d)", h_pos, v_pos);
	    h_pos <= 0;
	    v_pos <= v_pos + 1;
	 end // if (!hv_rgb[1] && prev_hv_rgb[1]

	 // Otherwise..
	 if(hv_rgb[1]) begin
//	    if(!prev_hv_rgb[1])
//	       $display("Display active (%d,%d)", h_pos, v_pos);
	    h_pos <= h_pos + 1;
	 end // if (hv_rgb[1])
	 
      end

      prev_hv_rgb <= hv_rgb;
   end


   // RGB color registers for image composition logic
   reg [5:0] rgb_out;
   reg [6:0] sprite0_rgba_out, sprite1_rgba_out,
   	     sprite2_rgba_out, sprite3_rgba_out;
   reg bg_display_bit;


   // Game state registers live in the context of this module
   reg [5:0] 	maze_bg, maze_fg;
   reg [5:0] 	sprite0_cfg, sprite1_cfg, sprite2_cfg, sprite3_cfg;
   reg [9:0] 	sprite0_x, sprite0_y, sprite1_x, sprite1_y,
   		sprite2_x, sprite2_y, sprite3_x, sprite3_y;

   // Game state registers prime (loaded into the real ones at every vblank)
   reg [5:0] 	maze_bgp, maze_fgp;
   reg [5:0] 	sprite0_cfgp, sprite1_cfgp, sprite2_cfgp, sprite3_cfgp;
   reg [4:0]   	sprite0_xsubp, sprite0_ysubp, sprite1_xsubp, sprite1_ysubp,
    		sprite2_xsubp, sprite2_ysubp, sprite3_xsubp, sprite3_ysubp,
		sprite0_xsuperp, sprite0_ysuperp, sprite1_xsuperp, sprite1_ysuperp,
		sprite2_xsuperp, sprite2_ysuperp, sprite3_xsuperp, sprite3_ysuperp;

   wire [9:0] 	sprite0_xp, sprite0_yp, sprite1_xp, sprite1_yp,
   	        sprite2_xp, sprite2_yp, sprite3_xp, sprite3_yp;
   assign sprite0_xp = { sprite0_xsuperp, sprite0_xsubp };
   assign sprite0_yp = { sprite0_ysuperp, sprite0_ysubp };
   assign sprite1_xp = { sprite1_xsuperp, sprite1_xsubp };
   assign sprite1_yp = { sprite1_xsuperp, sprite1_ysubp };
   assign sprite2_xp = { sprite2_xsuperp, sprite2_xsubp };
   assign sprite2_yp = { sprite2_xsuperp, sprite2_ysubp };
   assign sprite3_xp = { sprite3_xsuperp, sprite3_xsubp };
   assign sprite3_yp = { sprite3_xsuperp, sprite3_ysubp };


   // Processor <-> Game State interface
   reg [4:0] 	reg_fill_counter;
   reg 		lcfg_we;
   reg 		hard_reset_observed;

   // Add status registers as status bits show up
   // maybe an hblank_active one?  Who knows.
   assign status_read[0] = hard_reset_observed;
   assign status_read[1] = 0;
   assign status_read[2] = 0;
   assign status_read[3] = 0;
   assign status_read[4] = 0;
   assign status_read[5] = 0;
   assign status_read[6] = 0;
   assign status_read[7] = 0;

   // Writeback loop
   // Load game config registers with the prime registers
   always @(posedge clk) begin
      
      if(reset) begin
	 // Set immediate defaults (for testing)
	 maze_bg <= 6'b000000;
	 maze_fg <= 6'b000000;
	 sprite0_cfg <= 0;  // test image idx 2
	 sprite0_x <= 0;
	 sprite0_y <= 480;
	 sprite1_cfg <= 0;
	 sprite1_x <= 0;
	 sprite1_y <= 480;
	 sprite2_cfg <= 0;
	 sprite2_x <= 0;
	 sprite2_y <= 480;
	 sprite3_cfg <= 0;
	 sprite3_x <= 0;
	 sprite3_y <= 480;

      end else begin // if (reset)

	 if(positive_vblank) begin
	    // Copy the game state prime registers into the real ones
	    // The game processor gets the same positive_vblank spike,
	    // let's just hope it finished its job.
	    maze_bg <= maze_bgp;
	    maze_fg <= maze_fgp;
	    sprite0_cfg <= sprite0_cfgp;
	    sprite0_x <= sprite0_xp;
	    sprite0_y <= sprite0_yp;
	    sprite1_cfg <= sprite1_cfgp;
	    sprite1_x <= sprite1_xp;
	    sprite1_y <= sprite1_yp;
	    sprite2_cfg <= sprite2_cfgp;
	    sprite2_x <= sprite2_xp;
	    sprite2_y <= sprite2_yp;
	    sprite3_cfg <= sprite3_cfgp;
	    sprite3_x <= sprite3_xp;
	    sprite3_y <= sprite3_yp;

	 end // if (positive_vblank)

      end // else: !if(reset)

   end // always @ (posedge clk)

   
   // The CPU can write to all configuration registers by sequentially
   // writing to cfg_write and strobing config_we.
   always @(posedge clk) begin

      if(reset) begin

	 hard_reset_observed <= 1;
	 reg_fill_counter <= 0;

	 // Maze colors default to black on black
	 maze_bgp <= 6'b000000;
	 maze_fgp <= 6'b000000;

	 // Sprites default to locations off the screen
	 sprite0_cfgp <= 0;
	 sprite0_xsuperp <= 0;
	 sprite0_xsubp <= 0;
	 sprite0_ysuperp <= 15;
	 sprite0_ysubp <= 0;
	 sprite1_cfgp <= 0;
	 sprite1_xsuperp <= 0;
	 sprite1_xsubp <= 0;
	 sprite1_ysuperp <= 15;
	 sprite1_ysubp <= 0;
	 sprite2_cfgp <= 0;
	 sprite2_xsuperp <= 0;
	 sprite2_xsubp <= 0;
	 sprite2_ysuperp <= 15;
	 sprite2_ysubp <= 0;
	 sprite3_cfgp <= 0;
	 sprite3_xsuperp <= 0;
	 sprite3_xsubp <= 0;
	 sprite3_ysuperp <= 15;
	 sprite3_ysubp <= 0;

      end else begin // if (reset)
	 if(cfg_we && !lcfg_we) begin

	    case(reg_fill_counter)
	      // BG/FG colors
	      // Useful for interesting palette effects or disabling
	      // the maze display unit when it isn't set up.
	      0: begin
		 maze_fgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      1: begin
		 maze_bgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end

	      // Sprite 0 settings
	      2: begin
		 sprite0_cfgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      3: begin
		 sprite0_xsuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      4: begin
		 sprite0_xsubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      5: begin
		 sprite0_ysuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      6: begin
		 sprite0_ysubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end

	      // Sprite 1 settings
	      7: begin
		 sprite1_cfgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      8: begin
		 sprite1_xsuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      9: begin
		 sprite1_xsubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      10: begin
		 sprite1_ysuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      11: begin
		 sprite1_ysubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end

	      // Sprite 2 settings
	      12: begin
		 sprite2_cfgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      13: begin
		 sprite2_xsuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      14: begin
		 sprite2_xsubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      15: begin
		 sprite2_ysuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      16: begin
		 sprite2_ysubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end

	      // Sprite 3 settings
	      17: begin
		 sprite3_cfgp <= cfg_write[5:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      18: begin
		 sprite3_xsuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      19: begin
		 sprite3_xsubp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      20: begin
		 sprite3_ysuperp <= cfg_write[4:0];
		 reg_fill_counter <= reg_fill_counter + 1;
	      end
	      21: begin
		 sprite3_ysubp <= cfg_write[4:0];
		 reg_fill_counter <= 0;
	      end
	    endcase

	    // hard reset is cleared when the config registers are written
	    hard_reset_observed <= 0;

	 end // if (write_go && !lwrite_go)
	 
      end // else: !if(reset)

      lcfg_we <= cfg_we;
   end


   
   // Sprites
   // sprite0: pac-man

   wire [6:0] pacman_rgba_out;
   pacman_sprite_unit local_pacman(clk, h_pos, v_pos,
				   sprite0_cfg, sprite0_x, sprite0_y,
				   pacman_rgba_out);
   always @(posedge clk)
      sprite0_rgba_out <= pacman_rgba_out;


   // sprite1: magenta ghost
   wire [6:0] ghost0_rgba_out;
   ghost0_sprite_unit #(7'b1100111)
          local_ghost0(clk, h_pos, v_pos,
		       sprite1_cfg, sprite1_x, sprite1_y,
		       ghost0_rgba_out);
   always @(posedge clk)
      sprite1_rgba_out <= ghost0_rgba_out;


   // sprite2: red ghost
   wire [6:0] ghost1_rgba_out;
   ghost0_sprite_unit #(7'b1100001)
          local_ghost1(clk, h_pos, v_pos,
		       sprite2_cfg, sprite2_x, sprite2_y,
		       ghost1_rgba_out);
   always @(posedge clk)
      sprite2_rgba_out <= ghost1_rgba_out;

   // sprite3: cyan ghost
   wire [6:0] ghost2_rgba_out;
   ghost0_sprite_unit #(7'b0011111)
          local_ghost2(clk, h_pos, v_pos,
		       sprite3_cfg, sprite3_x, sprite3_y,
		       ghost2_rgba_out);
   always @(posedge clk)
      sprite3_rgba_out <= ghost2_rgba_out;

   
   // Maze display unit
   // This acts as the background image
   wire       maze_display_bit;
   maze_display_unit local_maze_display(clk, h_pos, v_pos,
					maccess_addr[7:4], maccess_addr[3:0], maccess_read,
					maccess_we, maccess_write,
					maze_display_bit);

   // RGB output
   always @(posedge clk)
      bg_display_bit <= maze_display_bit;

   // Layer composition logic
   always @(posedge clk) begin
      if(end_rgb_enable) begin
         // Sprites drawn in reverse order
         if(sprite3_rgba_out[0]) begin
	    rgb_out <= sprite3_rgba_out[6:1];

         end else if(sprite2_rgba_out[0]) begin
	    rgb_out <= sprite2_rgba_out[6:1];

         end else if(sprite1_rgba_out[0]) begin
	    rgb_out <= sprite1_rgba_out[6:1];

         // Sprite 0 is on the bottom and goes last
         end else if(sprite0_rgba_out[0]) begin
	    rgb_out <= sprite0_rgba_out[6:1];

         // Maze on bottom
         end else
	    rgb_out <= bg_display_bit ? maze_fg : maze_bg;
      end else
         rgb_out <= 6'b000000;
   end
endmodule // vga_driver


// maze_display_unit
//
// This unit contains a mazekeeper unit, (see mazekeeper.v) which acts
// as its video memory, and a tile ROM with an image for each possible
// room state of the maze.
// The tile ROM is 1 BPP, as is its display_bit output.  Ideally this should
// look up a color in a palette.
//
// The size given to the maze unit is 16x16, meaning there are four row bits
// and four column bits.  This makes it easy to address using an 8-bit
// microprocessor core.
//
// Access to the maze RAM is provided through the maccess_ lines.
//
// maze_display_unit requires one clock, but likely has relatively long setup
// and propagation times.

module maze_display_unit(clk, h_pos, v_pos,
			 maccess_row, maccess_col, maccess_value,
			 maccess_we, maccess_data,
			 display_bit);

   input     clk;
   input [9:0] h_pos, v_pos;

   input [3:0] maccess_row, maccess_col;
   output [3:0] maccess_value;
   input [3:0] maccess_we, maccess_data;
   output      display_bit;
   
   reg [4:0] 		tile_x, tile_y;
   wire [3:0] 		tileidx; 		
   wire 		x_rangei;
   reg 			x_range;
   wire [6:0] 		x_coord;

   assign x_rangei = (h_pos[9:6] == 0) ? 0 :
		     ((h_pos[9:5] >= 18) ? 0 : 1);
   assign x_coord = (h_pos[9:5] - 2);

   // Instantiation of the maze keeper
   // Wire the maccess lines directly into it.
   // 20 wide, 15 high, 0:19=10100=5 bits, 0:14=1110=4 bits.
   mazekeeper #(16, 15, 4, 4)
      mk(clk, v_pos[8:5], x_coord[3:0], tileidx,
	 maccess_row, maccess_col, maccess_value,
	 maccess_we, maccess_data);

   // NOTE: mazekeeper latches tileidx, maccess_value.
   
   // The tile ROM is organized in a 4x4 grid of tiles of size 32x32.
   // (=16 kbits)  KEEP IT SMALL PLZ!
   // The 2 MSBs of the X grid coordinate represent left,right connectivity,
   // The 2 MSBs of the Y grid coordinate represent top,bottom connectivity.
   wire [6:0] 		hack_tile_x, hack_tile_y;
   wire 		tile_bit;
   
   assign hack_tile_x = { tileidx[1], tileidx[3], tile_x };
   assign hack_tile_y = { tileidx[0], tileidx[2], tile_y };

   rom_mazetiles local_tiles(hack_tile_x, hack_tile_y, tile_bit);

   assign display_bit = (x_range && tile_bit);
   
   always @(posedge clk) begin
      tile_x <= h_pos[4:0];
      tile_y <= v_pos[4:0];
      x_range <= x_rangei;
   end

endmodule // maze_display_unit


// Pac-Man character sprite unit
// Outputs bits from its image ROM at the correct locations
// on the screen.  Merging with other screen layers happens upstream.
//
// Image configuration bit meanings:
// bit 0: X mirror
// bit 1: Y mirror
// bit 2: X/Y exchange (rotates 90 degrees)
// bits [5:3]: image index selection

module pacman_sprite_unit(clk, h_pos, v_pos,
			  cfg_bitmap, cfg_x, cfg_y,
			  rgba_out);
   parameter SIZE_X = 32,
	     SIZE_Y = 32,
	     BPP = 7,
	     CFG_BITS = 6,
	     BASE_COLOR = 7'b1111001;
   
   input clk;
   input [9:0] h_pos, v_pos;
   input [CFG_BITS-1:0] cfg_bitmap;
   input [9:0] cfg_x, cfg_y;
   output [BPP-1:0] rgba_out;

   reg [10:0] 	    x_sub, y_sub;
   reg [CFG_BITS-1:0] cfg_bitmapi;
   

   always @(posedge clk) begin
      x_sub <= (h_pos - cfg_x);
      y_sub <= (v_pos - cfg_y);
      cfg_bitmapi <= cfg_bitmap;
   end // always @ (posedge clk)


   wire [4:0] 	  x_rompos_1, y_rompos_1, x_rompos_2, y_rompos_2;

   // Handle inversions and reversals
   // hey, negating & muxing 5 bits is probably faster than subtracting 10.
   // ORDER: first swap, then negate.  A lot easier to understand
   // from a software point of view.
   assign x_rompos_1 = (cfg_bitmapi[2] ? y_sub : x_sub);
   assign y_rompos_1 = (cfg_bitmapi[2] ? x_sub : y_sub);
   assign x_rompos_2 = (cfg_bitmapi[0] ? (-x_rompos_1[4:0]) : x_rompos_1[4:0]);
   assign y_rompos_2 = (cfg_bitmapi[1] ? (-y_rompos_1[4:0]) : y_rompos_1[4:0]);

   // Final rom index composition
   wire [6:0] 	  x_rompos;
   wire [4:0] 	  y_rompos;
   assign x_rompos = { cfg_bitmapi[4:3], x_rompos_2 };
   assign y_rompos = y_rompos_2;

   wire 	  pacman_bit;
   rom_pacman myrom(x_rompos, y_rompos, pacman_bit);

   // Set the output to be:
   // Yellow if the ROM returns 1,
   // Opaque if the ROM returns 1,
   // Black and transparent otherwise.
   wire [BPP-1:0] rom_rgba;
   assign rom_rgba = pacman_bit ? BASE_COLOR : 7'b0000000;

   wire       x_range;
   wire       y_range;
   assign x_range = (x_sub < SIZE_X);
   assign y_range = (y_sub < SIZE_Y);
   assign rgba_out = (x_range && y_range) ? rom_rgba : 7'b0000000;

endmodule // pacman_sprite_unit


// ghost zero sprite unit
// works a whole lot like pac-man.
module ghost0_sprite_unit(clk, h_pos, v_pos,
			  cfg_bitmap, cfg_x, cfg_y,
			  rgba_out);
   parameter BASE_COLOR = 7'b0000000,
   	     SIZE_X = 32,
	     SIZE_Y = 32,
	     BPP = 7,
	     CFG_BITS = 6;
   
   input clk;
   input [9:0] h_pos, v_pos;
   input [CFG_BITS-1:0] cfg_bitmap;
   input [9:0] cfg_x, cfg_y;
   output [BPP-1:0] rgba_out;

   reg [10:0] 	    x_sub, y_sub;
   reg [CFG_BITS-1:0] cfg_bitmapi;

   always @(posedge clk) begin
      x_sub <= (h_pos - cfg_x);
      y_sub <= (v_pos - cfg_y);
      cfg_bitmapi <= cfg_bitmap;
   end // always @ (posedge clk)

   wire [4:0] 	  x_rompos_1, y_rompos_1, x_rompos_2, y_rompos_2;
   assign x_rompos_1 = (cfg_bitmapi[2] ? y_sub : x_sub);
   assign y_rompos_1 = (cfg_bitmapi[2] ? x_sub : y_sub);
   assign x_rompos_2 = (cfg_bitmapi[0] ? (-x_rompos_1[4:0]) : x_rompos_1[4:0]);
   assign y_rompos_2 = (cfg_bitmapi[1] ? (-y_rompos_1[4:0]) : y_rompos_1[4:0]);

   // Final rom index composition
   wire [6:0] 	  x_rompos;
   wire [4:0] 	  y_rompos;
   assign x_rompos = { cfg_bitmapi[4:3], x_rompos_2 };
   assign y_rompos = y_rompos_2;

   wire [1:0] ghost_bits;
   // Temporary hack: NEED GHOST IMAGE ROM
   rom_pacman myrom(x_rompos, y_rompos, ghost_bits[1]);
   assign ghost_bits[0] = 0;

   // ghost_bits:
   // 00 = transparent black
   // 01 = opaque black
   // 10 = pink/red/ghost color
   // 11 = white
   wire [BPP-1:0] rom_rgba;
   assign rom_rgba = ghost_bits[1] ? (ghost_bits[0] ? 7'b1111111 : BASE_COLOR)
                     : (ghost_bits[0] ? 7'b0000001 : 7'b0000000);

   wire       x_range;
   wire       y_range;
   assign x_range = (x_sub < SIZE_X);
   assign y_range = (y_sub < SIZE_Y);
   assign rgba_out = (x_range && y_range) ? rom_rgba : 7'b0000000;

endmodule


`include "hv_rgb.v"
`include "mazekeeper.v"
`include "mazetiles_x.v"
`include "pacman.v"

