module vga_driver(clk, reset,
		  h_sync, v_sync, h_rgb, v_rgb,
		  kbd_data, kbd_int,
		  h_sync_out, v_sync_out, rgb_out);
   input clk, reset;
   input h_sync, v_sync, h_rgb, v_rgb;
   input [7:0] kbd_data;
   input       kbd_int;
   output      h_sync_out, v_sync_out;
   output [5:0] rgb_out;


   // Sync signal pipeline/delay
   // NOTE: make sure this is *as deep* as the pixel value pipeline.
   reg [1:0] 	hv_sync0, hv_sync1, hv_sync2;
   assign h_sync_out = hv_sync2[1];
   assign v_sync_out = hv_sync2[0];

   always @(posedge clk) begin
      hv_sync0 <= {h_sync, v_sync};
      hv_sync1 <= hv_sync0;
      hv_sync2 <= hv_sync1;
   end // always @ (posedge clk)


   // Screen position counter functionality
   reg [9:0] 	h_pos, v_pos;
   reg [1:0] 	prev_hv_rgb;
   wire [1:0] 	hv_rgb;
   wire [1:0] 	hv_blank;

   parameter 	HV_VBLANK = 2'b00, HV_HBLANK = 2'b01, HV_ACTIVE = 2'b11;

   assign hv_rgb = {h_rgb, v_rgb};
   // hv_blank is for texture units below and operates on t-1 time.
   assign hv_blank = { (prev_hv_rgb == HV_HBLANK),
		       (prev_hv_rgb == HV_VBLANK) };

   // horizontal and vertical position drivers   
   always @(posedge clk) begin
      case(hv_rgb)
	HV_VBLANK: begin
	   h_pos <= 0;
	   v_pos <= 0;
	end // case: HV_VBLANK

	HV_ACTIVE: begin
	   h_pos <= h_pos + 1;
	   $display("display active, x=%h y=%h", h_pos, v_pos);
	end // case: HV_ACTIVE

	HV_HBLANK: begin
	   if(prev_hv_rgb == HV_ACTIVE) begin
	      h_pos <= 0;
	      v_pos <= v_pos + 1;
	   end // if (prev_hv_rgb)
	end // case: HV_HBLANK

      endcase // case(hvstate)
      prev_hv_rgb <= hv_rgb;
   end


   // Game control unit
   // Self-contains the game state registers.
   wire [9:0] lh_scroll, lv_scroll;
   vga_game_control local_gamecontrol(clk, reset, hv_blank, kbd_data, kbd_irq,
				      lh_scroll, lv_scroll);
   

   // Background texture unit
   reg [5:0] bg_rgb_out;

   wire [5:0] bg_rgb_latchme;
   vga_background_unit local_background(clk, reset, h_pos, v_pos,
					lh_scroll, lv_scroll,
					hv_blank, bg_rgb_latchme);
   always @(posedge clk) begin
      bg_rgb_out <= bg_rgb_latchme;
   end // always @ (posedge clk)

   assign rgb_out = bg_rgb_out;
   //reg [5:0] 	rgb_out;
   
endmodule // vga_driver


module vga_game_control(clk, reset, hv_blank,
			kbd_data, kbd_irq,
			h_scroll, v_scroll);
   input clk, reset;
   input [1:0] hv_blank;
   input [7:0] kbd_data;
   input       kbd_irq;
   output [9:0] h_scroll, v_scroll;

   reg [1:0]	last_hv_blank;
   wire 	positive_vblank = (hv_blank[0] && !last_hv_blank[0]);
   

   // Game state registers live in the context of this module
   reg [9:0] 	h_scroll, v_scroll;

   always @(posedge clk) begin
      
      if(reset) begin
	 h_scroll <= 0;
	 v_scroll <= 0;
      end else begin // if (reset)

	 if(positive_vblank) begin
	    if(kbd_data[5] && kbd_data[4])
	       h_scroll <= h_scroll + 5;
	    if(kbd_data[5] && !kbd_data[4])
	       h_scroll <= h_scroll - 5;
	 end // if (positive_vblank)

      end // else: !if(reset)

      last_hv_blank <= hv_blank;
   end // always @ (posedge clk)
   
endmodule


module vga_background_unit(clk, reset, h_pos, v_pos, ih_scroll, iv_scroll,
			   hv_blank, rgb_out);
   input clk, reset;
   input [9:0] h_pos, v_pos;
   input [9:0] ih_scroll, iv_scroll;
   input       hv_blank;
   output [5:0]  rgb_out;

   wire [10:0] 	 h_scroll, v_scroll;
   wire [4:0] 	 itileidx;
   reg [4:0] 	 tileidx;
   reg [4:0] 	 tile_x, tile_y;

   assign h_scroll = (h_pos + ih_scroll);
   assign v_scroll = (v_pos + iv_scroll);
   rom_background_tilemap local_tilemap(h_scroll[10:5], v_scroll[10:5],
					itileidx);

   // The tile ROM is vertically oriented and repeats in the Y direction.
   // Therefore we stick the tile index in the upper 5 bits of y.
   wire [9:0]  hack_tile_y = { tileidx, tile_y };
   rom_tiles local_tiles(tile_x, hack_tile_y, rgb_out);
   
   always @(posedge clk) begin
      tileidx <= itileidx;
      tile_x <= h_scroll[4:0];
      tile_y <= v_scroll[4:0];
   end

endmodule // vga_background_unit


module rom_background_tilemap(h_pos, v_pos, tileidx);
   input [5:0] h_pos, v_pos;
   output [4:0] tileidx;

   wire [5:0] 	intsum;
   
   // this is my ROM, hehehe
   assign intsum = (v_pos[4:0] + h_pos[4:0] );
   assign tileidx = intsum[4:0];
endmodule // rom_background_tilemap


